@@ -8,6 +8,7 @@ from typing import (
     List,
     MutableMapping,
     Optional,
+    Set,
     TypeVar,
 )
 from datetime import datetime
@@ -34,6 +35,7 @@ class PSLModel:
         rules: RuleContainer,
         temp_dir: Optional[str] = None,
         ignore_predicates_not_in_rules: bool = False,
+        required_predicates: Optional[Set[str]] = None,
     ):
         self.model = Model("psl-model")
         self.rules = rules
@@ -47,8 +49,14 @@ class PSLModel:
         assert not Path(self.temp_dir).exists(), f"{self.temp_dir} already exists"
 
         if ignore_predicates_not_in_rules:
+            if required_predicates is None:
+                required_predicates = set()
             using_predicates = []
             for p in predicates:
+                if p.name() in required_predicates:
+                    using_predicates.append(p)
+                    continue
+
                 pname = p.name() + "("
                 subrules = [
                     r
@@ -173,11 +181,23 @@ class PSLModel:
                 )
                 break
             except ModelError as e:
+                # error related to parser (not jvm), no retry
+                # non_recoverable_error = any(
+                #     line.find(k) != -1
+                #     for line in self.logs
+                #     for k in [
+                #         "org.linqs.psl.parser.antlr.PSLParser",
+                #         "Unique index or primary key violation",
+                #     ]
+                # )
+                non_recoverable_error = False
                 self.log_errors(e)
-                if not retry:
+                if not retry or non_recoverable_error:
                     raise
 
             logger.info("Retry PSL inference...")
+        else:
+            raise Exception("PSL inference failed")
 
         if cleanup_tempdir:
             shutil.rmtree(self.temp_dir)
@@ -214,9 +234,11 @@ class PSLModel:
                     )
                 df.to_csv(outdir / f"{pname}.{partition}.csv", index=False)
 
-    def log_errors(self, error: Exception):
+    def log_errors(self, error: Exception, clear: bool = True):
         logger.error(str(error))
         print("\n".join(self.logs))
+        if clear:
+            self.logs = []
 
     @staticmethod
     def normalize_probs(
@@ -357,7 +379,7 @@ class RuleContainer(MutableMapping[str, Rule]):
         return key in self.rules
 
     def __setitem__(self, key: str, value: Rule):
-        assert key not in self.rules
+        assert key not in self.rules, key
         self.rules[key] = value
 
     def __getitem__(self, key: str) -> Rule:
