@@ -1,27 +1,28 @@
 import itertools
 from enum import IntEnum
-from typing import Dict, List, Set
+from typing import Dict, List, Mapping, Optional, Set, Tuple
 from uuid import uuid4
+from grams.algorithm.candidate_graph.cg_graph import (
+    CGColumnNode,
+    CGEntityValueNode,
+    CGGraph,
+    CGLiteralValueNode,
+    CGStatementNode,
+)
 
 import networkx as nx
 import numpy as np
 import sm.misc as M
 import sm.outputs as O
-from grams.algorithm.semantic_graph import (
-    SGColumnNode,
-    SGEdge,
-    SGEntityValueNode,
-    SGLiteralValueNode,
-    SGNode,
-    SGStatementNode,
-)
-from grams.algorithm.wdont import WDOnt
+
 from grams.inputs.linked_table import LinkedTable
-from kgdata.wikidata.models import QNode, WDClass, WDProperty
+from kgdata.wikidata.models import WDEntity, WDClass, WDProperty
 from rdflib import RDFS
 from sm.evaluation import sm_metrics
 from sm.misc import identity_func
 from sm.misc.graph import viz_graph
+from sm.outputs.semantic_model import ClassNode
+from sm.namespaces import WikidataNamespace
 
 
 class SMNodeType(IntEnum):
@@ -32,29 +33,45 @@ class SMNodeType(IntEnum):
     Literal = 4
 
 
-class WikidataSemanticModelHelper(WDOnt):
+class WikidataSemanticModelHelper:
     ENTITY_ID = "Q35120"
     ENTITY_LABEL = "Entity (Q35120)"
     ID_PROPS = {str(RDFS.label)}
 
+    def __init__(
+        self,
+        wdentities: Mapping[str, WDEntity],
+        wdentity_labels: Mapping[str, str],
+        wdclasses: Mapping[str, WDClass],
+        wdprops: Mapping[str, WDProperty],
+    ):
+        self.wdentities = wdentities
+        self.wdentity_labels = wdentity_labels
+        self.wdclasses = wdclasses
+        self.wdprops = wdprops
+        self.wdns = WikidataNamespace.create()
+
     def norm_sm(self, sm: O.SemanticModel):
         """ "Normalize the semantic model with the following modifications:
         1. Add readable label to edge and class
         2. Convert direct link (without statement) to have statement except the id props.
         """
-        new_sm = sm.clone()
+        new_sm = sm.deep_copy()
+        wdns = self.wdns
 
         # update readable label
         for n in new_sm.iter_nodes():
             if isinstance(n, O.ClassNode):
-                if self.is_uri_qnode(n.abs_uri):
-                    n.readable_label = self.get_qnode_label(n.abs_uri)
+                if wdns.is_abs_uri_qnode(n.abs_uri):
+                    n.readable_label = self.get_qnode_label(
+                        wdns.get_entity_id(n.abs_uri)
+                    )
             elif isinstance(n, O.LiteralNode):
-                if self.is_uri_qnode(n.value):
-                    n.readable_label = self.get_qnode_label(n.value)
+                if wdns.is_abs_uri_qnode(n.value):
+                    n.readable_label = self.get_qnode_label(wdns.get_entity_id(n.value))
         for e in new_sm.iter_edges():
             if e.abs_uri not in self.ID_PROPS:
-                e.readable_label = self.get_pnode_label(e.abs_uri)
+                e.readable_label = self.get_pnode_label(wdns.get_prop_id(e.abs_uri))
 
         # convert direct link
         for edge in list(new_sm.iter_edges()):
@@ -65,20 +82,17 @@ class WikidataSemanticModelHelper(WDOnt):
 
             if (
                 not isinstance(source, O.ClassNode)
-                or source.abs_uri != WDOnt.STATEMENT_URI
+                or source.abs_uri != WikidataNamespace.STATEMENT_URI
             ) and (
                 not isinstance(target, O.ClassNode)
-                or target.abs_uri != WDOnt.STATEMENT_URI
+                or target.abs_uri != WikidataNamespace.STATEMENT_URI
             ):
                 # this is direct link, we replace its edge
                 assert len(new_sm.get_edges_between_nodes(source.id, target.id)) == 1
                 new_sm.remove_edges_between_nodes(source.id, target.id)
                 stmt = O.ClassNode(
-                    str(uuid4()),
-                    WDOnt.STATEMENT_URI,
-                    WDOnt.STATEMENT_REL_URI,
-                    False,
-                    "Statement",
+                    abs_uri=WikidataNamespace.STATEMENT_URI,
+                    rel_uri=wdns.get_rel_uri(WikidataNamespace.STATEMENT_URI),
                 )
                 new_sm.add_node(stmt)
 
@@ -109,12 +123,15 @@ class WikidataSemanticModelHelper(WDOnt):
         """This is a reverse function of `norm_sm`:
         1. Remove an intermediate statement if it doesn't have any qualifiers
         """
-        new_sm = sm.clone()
+        new_sm = sm.copy()
 
         for n in sm.iter_nodes():
-            if isinstance(n, O.ClassNode) and n.abs_uri == WDOnt.STATEMENT_URI:
-                inedges = sm.incoming_edges(n.id)
-                outedges = sm.outgoing_edges(n.id)
+            if (
+                isinstance(n, O.ClassNode)
+                and n.abs_uri == WikidataNamespace.STATEMENT_URI
+            ):
+                inedges = sm.in_edges(n.id)
+                outedges = sm.out_edges(n.id)
                 if len(outedges) == 1 and outedges[0].abs_uri == inedges[0].abs_uri:
                     # no qualifiers
                     new_sm.remove_node(n.id)
@@ -133,15 +150,13 @@ class WikidataSemanticModelHelper(WDOnt):
                         )
         return new_sm
 
-    def create_sm(self, table: LinkedTable, cpa: nx.MultiDiGraph, cta: Dict[int, str]):
+    def create_sm(self, table: LinkedTable, cpa: CGGraph, cta: Dict[int, str]):
         """Create a semantic model from outputs of CPA and CTA tasks"""
         sm = O.SemanticModel()
-        # create class nodes first
-        classcount = {}
         classmap = {}  # mapping from column to its class node
+        wdns = self.wdns
         for cid, qnode_id in cta.items():
             dnode = O.DataNode(
-                id=f"col-{cid}",
                 col_index=cid,
                 label=table.table.columns[cid].name or "",
             )
@@ -149,60 +164,51 @@ class WikidataSemanticModelHelper(WDOnt):
             # somehow, they may end-up predict multiple classes, we need to select one
             if qnode_id.find(" ") != -1:
                 qnode_id = qnode_id.split(" ")[0]
-            curl = self.get_qnode_uri(qnode_id)
-            cnode_id = f"{curl}:{classcount.get(qnode_id, 0)}"
-            classcount[qnode_id] = classcount.get(qnode_id, 0) + 1
+            curl = wdns.get_entity_abs_uri(qnode_id)
 
             try:
-                cnode_label = self.get_qnode_label(curl)
+                cnode_label = self.get_qnode_label(qnode_id)
             except KeyError:
-                cnode_label = f"wd:{qnode_id}"
+                cnode_label = wdns.get_entity_rel_uri(qnode_id)
             cnode = O.ClassNode(
-                id=cnode_id,
                 abs_uri=curl,
-                rel_uri=f"wd:{qnode_id}",
+                rel_uri=wdns.get_entity_rel_uri(qnode_id),
                 readable_label=cnode_label,
             )
-            classmap[dnode.id] = cnode.id
-
             sm.add_node(dnode)
             sm.add_node(cnode)
+            classmap[dnode.col_index] = cnode.id
             sm.add_edge(
                 O.Edge(
                     source=cnode.id,
                     target=dnode.id,
                     abs_uri=str(RDFS.label),
-                    rel_uri="rdfs:label",
+                    rel_uri=wdns.get_rel_uri(RDFS.label),
                 )
             )
 
         # do a final sweep to add subject columns that are not in CTA
-        for uid, unode in cpa.nodes(data="data"):  # type: ignore
-            unode: SGNode
-            if not isinstance(unode, SGColumnNode):
+        for unode in cpa.iter_nodes():
+            if not isinstance(unode, CGColumnNode):
                 continue
-            outdegree: int = cpa.out_degree(uid)  # type: ignore
-            if outdegree > 0 and not sm.has_node(f"col-{unode.column}"):
+            outdegree: int = cpa.out_degree(unode.id)
+            if outdegree > 0 and not sm.has_data_node(unode.column):
                 # add data node to the graph and use the entity class (all instances belong to this class) to describe this data node
                 dnode = O.DataNode(
-                    id=f"col-{unode.column}",
                     col_index=unode.column,
                     label=table.table.columns[unode.column].name or "",
                 )
                 sm.add_node(dnode)
 
-                curl = self.get_qnode_uri(self.ENTITY_ID)
-                cnode_id = f"{curl}:{classcount.get(self.ENTITY_ID, 0)}"
-                classcount[self.ENTITY_ID] = classcount.get(self.ENTITY_ID, 0) + 1
-                sm.add_node(
+                curl = wdns.get_entity_abs_uri(self.ENTITY_ID)
+                cnode_id = sm.add_node(
                     O.ClassNode(
-                        id=cnode_id,
                         abs_uri=curl,
-                        rel_uri=f"wd:{self.ENTITY_ID}",
+                        rel_uri=wdns.get_entity_rel_uri(self.ENTITY_ID),
                         readable_label=self.ENTITY_LABEL,
                     )
                 )
-                classmap[dnode.id] = cnode_id
+                classmap[dnode.col_index] = cnode_id
 
                 sm.add_edge(
                     O.Edge(
@@ -214,83 +220,92 @@ class WikidataSemanticModelHelper(WDOnt):
                 )
 
         # now add remaining edges and remember to use class node instead of data node
-        for uid, vid, edge in cpa.edges(data="data"):  # type: ignore
-            edge: SGEdge
-            unode: SGNode = cpa.nodes[uid]["data"]
-            vnode: SGNode = cpa.nodes[vid]["data"]
+        cpa_idmap = {}
+        for edge in cpa.edges():
+            unode = cpa.get_node(edge.source)
+            vnode = cpa.get_node(edge.target)
 
-            if isinstance(unode, SGColumnNode):
+            if isinstance(unode, CGColumnNode):
                 # outgoing edge is from a class node instead of a data node
-                suid = classmap[f"col-{unode.column}"]
+                suid = classmap[unode.column]
                 source = sm.get_node(suid)
-            elif isinstance(unode, SGEntityValueNode):
-                source = O.LiteralNode(
-                    id=unode.id,
-                    value=self.get_qnode_uri(unode.qnode_id),
-                    readable_label=self.get_qnode_label(unode.qnode_id),
-                    datatype=O.LiteralNodeDataType.Entity,
-                    is_in_context=unode.qnode_id == table.context.page_entity_id,
-                )
-                sm.add_node(source)
+            elif isinstance(unode, CGEntityValueNode):
+                if unode.id not in cpa_idmap:
+                    source = O.LiteralNode(
+                        value=wdns.get_entity_abs_uri(unode.qnode_id),
+                        readable_label=self.get_qnode_label(unode.qnode_id),
+                        datatype=O.LiteralNodeDataType.Entity,
+                        is_in_context=unode.qnode_id == table.context.page_entity_id,
+                    )
+                    cpa_idmap[unode.id] = sm.add_node(source)
+                else:
+                    source = sm.get_node(cpa_idmap[unode.id])
             else:
                 assert isinstance(
-                    unode, SGStatementNode
+                    unode, CGStatementNode
                 ), "Outgoing edge can't not be from literal"
-                # create a statement node
-                source = O.ClassNode(
-                    id=unode.id,
-                    abs_uri=WDOnt.STATEMENT_URI,
-                    rel_uri=WDOnt.STATEMENT_REL_URI,
-                )
-                sm.add_node(source)
-
-            if isinstance(vnode, SGColumnNode):
-                svid = f"col-{vnode.column}"
-                if svid in classmap:
-                    target = sm.get_node(classmap[svid])
-                elif sm.has_node(svid):
-                    target = sm.get_node(svid)
+                if unode.id not in cpa_idmap:
+                    # create a statement node
+                    source = O.ClassNode(
+                        abs_uri=wdns.STATEMENT_URI,
+                        rel_uri=wdns.get_rel_uri(wdns.STATEMENT_URI),
+                    )
+                    cpa_idmap[unode.id] = sm.add_node(source)
                 else:
+                    source = sm.get_node(cpa_idmap[unode.id])
+
+            if isinstance(vnode, CGColumnNode):
+                if vnode.column in classmap:
+                    target = sm.get_node(classmap[vnode.column])
+                elif sm.has_data_node(vnode.column):
+                    target = sm.get_data_node(vnode.column)
+                elif vnode.id not in cpa_idmap:
                     target = O.DataNode(
-                        id=f"col-{vnode.column}",
                         col_index=vnode.column,
                         label=table.table.columns[vnode.column].name or "",
                     )
-                    sm.add_node(target)
-            elif isinstance(vnode, SGEntityValueNode):
-                target = O.LiteralNode(
-                    id=vnode.id,
-                    value=self.get_qnode_uri(vnode.qnode_id),
-                    readable_label=self.get_qnode_label(vnode.qnode_id),
-                    datatype=O.LiteralNodeDataType.Entity,
-                    is_in_context=vnode.qnode_id == table.context.page_entity_id,
-                )
-                sm.add_node(target)
-            elif isinstance(vnode, SGLiteralValueNode):
-                target = O.LiteralNode(
-                    id=vnode.id,
-                    value=vnode.value.to_string_repr(),
-                    readable_label=vnode.label,
-                    datatype=O.LiteralNodeDataType.String,
-                )
-                sm.add_node(target)
+                    cpa_idmap[vnode.id] = sm.add_node(target)
+                else:
+                    target = sm.get_node(cpa_idmap[vnode.id])
+            elif isinstance(vnode, CGEntityValueNode):
+                if vnode.id not in cpa_idmap:
+                    target = O.LiteralNode(
+                        value=wdns.get_entity_abs_uri(vnode.qnode_id),
+                        readable_label=self.get_qnode_label(vnode.qnode_id),
+                        datatype=O.LiteralNodeDataType.Entity,
+                        is_in_context=vnode.qnode_id == table.context.page_entity_id,
+                    )
+                    cpa_idmap[vnode.id] = sm.add_node(target)
+                else:
+                    target = sm.get_node(cpa_idmap[vnode.id])
+            elif isinstance(vnode, CGLiteralValueNode):
+                if vnode.id not in cpa_idmap:
+                    target = O.LiteralNode(
+                        value=vnode.value.to_string_repr(),
+                        readable_label=vnode.label,
+                        datatype=O.LiteralNodeDataType.String,
+                    )
+                    cpa_idmap[vnode.id] = sm.add_node(target)
+                else:
+                    target = sm.get_node(cpa_idmap[vnode.id])
             else:
-                # create a statement node
-                target = O.ClassNode(
-                    id=vnode.id,
-                    abs_uri=WDOnt.STATEMENT_URI,
-                    rel_uri=WDOnt.STATEMENT_REL_URI,
-                )
-                sm.add_node(target)
+                if vnode.id not in cpa_idmap:
+                    # create a statement node
+                    target = O.ClassNode(
+                        abs_uri=WikidataNamespace.STATEMENT_URI,
+                        rel_uri=wdns.get_rel_uri(WikidataNamespace.STATEMENT_URI),
+                    )
+                    cpa_idmap[vnode.id] = sm.add_node(target)
+                else:
+                    target = sm.get_node(cpa_idmap[vnode.id])
 
-            prop_uri = self.get_prop_uri(edge.predicate)
             sm.add_edge(
                 O.Edge(
                     source=source.id,
                     target=target.id,
-                    abs_uri=self.get_prop_uri(edge.predicate),
-                    rel_uri=f"p:{edge.predicate}",
-                    readable_label=self.get_pnode_label(prop_uri),
+                    abs_uri=wdns.get_prop_abs_uri(edge.predicate),
+                    rel_uri=wdns.get_prop_rel_uri(edge.predicate),
+                    readable_label=self.get_pnode_label(edge.predicate),
                 )
             )
 
@@ -302,42 +317,42 @@ class WikidataSemanticModelHelper(WDOnt):
         sm: O.SemanticModel,
         strict: bool = True,
         force_inversion: bool = False,
-        incorrect_invertible_props: Set[str] = None,
+        limited_invertible_props: Optional[Set[str]] = None,
+        incorrect_invertible_props: Optional[Set[str]] = None,
     ):
         """Given a semantic model (not being modified), generate equivalent models by inferring inverse properties.
 
+        Currently, we only inverse the properties, not qualifiers.
+
         Parameters
         ----------
         sm: the input semantic model (original)
         strict: whether to throw exception when target of an inverse property is not a class.
         force_inversion: only work when strict mode is set to false. Without force_inverse, we skip inverse properties,
                        otherwise, we generate an inverse model with a special class: wikibase:DummyClassForInversion
-
+        limited_invertible_props: if provided, only generate inverse properties for these properties.
+        incorrect_invertible_props: if provided, skip generating inverse properties for these properties.
         Returns
         -------
-
-        """
-        """Given an semantic model (not being modified), generate equivalent models
-        by inferring inverse properties. Running on strict mode mean it will check if the invertible property is apply
-        to a non-class node (column that .
-
-        Currently, we only inverse the properties, not qualifiers.
         """
         sm = self.norm_sm(sm)
+        wdns = self.wdns
+
         if incorrect_invertible_props is None:
             incorrect_invertible_props = set()
-        invertible_stmts = []
-        is_class_fn = lambda n1: n1.is_class_node or (
-            n1.is_literal_node and self.is_uri_qnode(n1.value)
+
+        invertible_stmts: List[O.ClassNode] = []
+        is_class_fn = lambda n1: isinstance(n1, O.ClassNode) or (
+            isinstance(n1, O.LiteralNode) and wdns.is_abs_uri_qnode(n1.value)
         )
 
         for n in sm.iter_nodes():
-            if n.is_class_node and WDOnt.is_uri_statement(n.abs_uri):
-                inedges = sm.incoming_edges(n.id)
-                outedges = sm.outgoing_edges(n.id)
+            if isinstance(n, O.ClassNode) and wdns.is_abs_uri_statement(n.abs_uri):
+                inedges = sm.in_edges(n.id)
+                outedges = sm.out_edges(n.id)
                 # only has one prop
                 (prop,) = list({inedge.abs_uri for inedge in inedges})
-                pid = self.get_prop_id(prop)
+                pid = wdns.get_prop_id(prop)
                 stmt_has_value = False
                 for outedge in outedges:
                     if outedge.abs_uri != prop:
@@ -350,6 +365,10 @@ class WikidataSemanticModelHelper(WDOnt):
                 if (
                     len(self.wdprops[pid].inverse_properties) > 0
                     and pid not in incorrect_invertible_props
+                    and (
+                        limited_invertible_props is None
+                        or pid in limited_invertible_props
+                    )
                     and stmt_has_value
                 ):
                     # invertible property
@@ -365,46 +384,56 @@ class WikidataSemanticModelHelper(WDOnt):
                     elif strict:
                         raise Exception(f"{pid} is not invertible")
                     elif force_inversion:
-                        assert sm.get_node(
-                            outedge.target
-                        ).is_data_node, "Clearly the model is wrong, you have an inverse property to a literal node"
+                        assert isinstance(
+                            sm.get_node(outedge.target), O.DataNode
+                        ), "Clearly the model is wrong, you have an inverse property to a literal node"
                         invertible_stmts.append(n)
 
         # we have N statement, so we are going to have N! - 1 ways. It's literally a cartesian product
         all_choices = []
         for stmt in invertible_stmts:
             # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
-            (inedge,) = sm.incoming_edges(stmt.id)
-            choice = [(stmt, None, None)]
+            (inedge,) = sm.in_edges(stmt.id)
+            choice: List[Tuple[O.ClassNode, Optional[str], Optional[str]]] = [
+                (stmt, None, None)
+            ]
             for invprop in self.wdprops[
-                self.get_prop_id(inedge.abs_uri)
+                wdns.get_prop_id(inedge.abs_uri)
             ].inverse_properties:
-                choice.append((stmt, self.get_prop_uri(invprop), f"p:{invprop}"))
+                choice.append(
+                    (
+                        stmt,
+                        wdns.get_prop_abs_uri(invprop),
+                        wdns.get_prop_rel_uri(invprop),
+                    )
+                )
             all_choices.append(choice)
 
         n_choices = np.prod([len(c) for c in all_choices]) - 1
         if n_choices > 256:
             raise sm_metrics.PermutationExplosion("Too many possible semantic models")
 
-        all_choices = list(itertools.product(*all_choices))
+        all_choices_perm: List[
+            Tuple[Tuple[O.ClassNode, Optional[str], Optional[str]]]
+        ] = list(itertools.product(*all_choices))
         assert all(
-            invprop is None for _, invprop, _ in all_choices[0]
+            invprop is None for _, invprop, _ in all_choices_perm[0]
         ), "First choice is always the current semantic model"
         new_sms = [sm]
-        for choice in all_choices[1:]:
-            new_sm = sm.clone()
+        for choice_perm in all_choices_perm[1:]:
+            new_sm = sm.copy()
             # we now change the statement from original prop to use the inverse prop (change direction)
             # if the invprop is not None
-            for stmt, invprop_abs_uri, invprop_rel_uri in choice:
-                if invprop_abs_uri is None:
+            for stmt, invprop_abs_uri, invprop_rel_uri in choice_perm:
+                if invprop_abs_uri is None or invprop_rel_uri is None:
                     continue
-                readable_label = self.get_pnode_label(invprop_abs_uri)
+                readable_label = self.get_pnode_label(wdns.get_prop_id(invprop_abs_uri))
                 # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
-                (inedge,) = sm.incoming_edges(stmt.id)
+                (inedge,) = sm.in_edges(stmt.id)
                 # statement must have only one property
                 (outedge,) = [
                     outedge
-                    for outedge in sm.outgoing_edges(stmt.id)
+                    for outedge in sm.out_edges(stmt.id)
                     if outedge.abs_uri == inedge.abs_uri
                 ]
                 assert (
@@ -417,11 +446,10 @@ class WikidataSemanticModelHelper(WDOnt):
 
                 target = sm.get_node(outedge.target)
                 if not is_class_fn(target):
-                    assert target.is_data_node
+                    assert isinstance(target, O.DataNode)
                     dummy_class_node = O.ClassNode(
-                        str(uuid4()),
-                        abs_uri="http://wikiba.se/ontology#DummyClassForInversion",
-                        rel_uri="wikibase:DummyClassForInversion",
+                        abs_uri=wdns.DUMMY_CLASS_FOR_INVERSION_URI,
+                        rel_uri=wdns.get_rel_uri(wdns.DUMMY_CLASS_FOR_INVERSION_URI),
                     )
                     new_sm.add_node(dummy_class_node)
                     new_sm.add_edge(
@@ -458,84 +486,11 @@ class WikidataSemanticModelHelper(WDOnt):
             new_sms.append(new_sm)
         return new_sms
 
-    def viz_sm(
-        self, sm: O.SemanticModel, outdir, graph_id
-    ) -> "WikidataSemanticModelHelper":
-        colors = {
-            SMNodeType.Column: dict(fill="#ffd666", stroke="#874d00"),
-            SMNodeType.Class: dict(fill="#b7eb8f", stroke="#135200"),
-            SMNodeType.Statement: dict(fill="#d9d9d9", stroke="#434343"),
-            SMNodeType.Entity: dict(fill="#C6E5FF", stroke="#5B8FF9"),
-            SMNodeType.Literal: dict(fill="#C6E5FF", stroke="#5B8FF9"),
-        }
-
-        def node_fn(uid, udata):
-            u: O.Node = udata["data"]
-            nodetype = self.get_node_type(u)
-            if isinstance(u, O.DataNode):
-                label = u.label
-            elif isinstance(u, O.ClassNode):
-                if u.abs_uri == WDOnt.STATEMENT_URI:
-                    label = ""
-                elif u.readable_label is None:
-                    if self.is_uri_qnode(u.abs_uri):
-                        label = self.get_qnode_label(u.abs_uri)
-                    else:
-                        label = u.label
-                else:
-                    label = u.readable_label
-            else:
-                label = u.readable_label
-
-            return {
-                "label": label,
-                "style": colors[nodetype],
-                "labelCfg": {
-                    "style": {
-                        "fill": "black",
-                        "background": {
-                            "padding": [4, 4, 4, 4],
-                            "radius": 3,
-                            **colors[nodetype],
-                        },
-                    }
-                },
-            }
-
-        def edge_fn(eid, edata):
-            edge: O.Edge = edata["data"]
-            label = edge.readable_label
-            if label is None:
-                if edge.abs_uri not in self.ID_PROPS:
-                    label = self.get_pnode_label(edge.abs_uri)
-                else:
-                    label = edge.label
-
-            return {"label": label}
-
-        viz_graph(sm.g, node_fn, edge_fn, outdir, graph_id)
-        return self
-
-    def get_node_type(self, n: O.Node):
-        if isinstance(n, O.LiteralNode):
-            if self.is_uri_qnode(n.value):
-                return SMNodeType.Entity
-            return SMNodeType.Literal
-        if isinstance(n, O.DataNode) or (
-            isinstance(n, O.ClassNode) and self.is_uri_column(n.abs_uri)
-        ):
-            return SMNodeType.Column
-        if isinstance(n, O.ClassNode):
-            if n.abs_uri == WDOnt.STATEMENT_URI:
-                return SMNodeType.Statement
-            return SMNodeType.Class
-        raise Exception("Unreachable!")
-
     def get_entity_columns(self, sm: O.SemanticModel) -> List[int]:
         ent_columns = []
         for dnode in sm.iter_nodes():
-            if dnode.is_data_node:
-                inedges = sm.incoming_edges(dnode.id)
+            if isinstance(dnode, O.DataNode):
+                inedges = sm.in_edges(dnode.id)
                 if len(inedges) == 0:
                     continue
                 assert len({edge.abs_uri for edge in inedges}) == 1, inedges
@@ -543,7 +498,9 @@ class WikidataSemanticModelHelper(WDOnt):
                 if edge_abs_uri in self.ID_PROPS:
                     assert len(inedges) == 1, inedges
                     source = sm.get_node(inedges[0].source)
-                    assert not self.is_uri_statement(source.abs_uri)
+                    assert isinstance(
+                        source, O.ClassNode
+                    ) and not self.wdns.is_abs_uri_statement(source.abs_uri)
                     ent_columns.append(dnode.col_index)
         return ent_columns
 
@@ -570,25 +527,28 @@ class WikidataSemanticModelHelper(WDOnt):
         """
         # norm the semantic model first
         sm = self.norm_sm(sm)
+        wdns = self.wdns
         schemas = {}
         for u in sm.iter_nodes():
-            if not u.is_class_node or WDOnt.is_uri_statement(u.abs_uri):
+            if not isinstance(u, O.ClassNode) or wdns.is_abs_uri_statement(u.abs_uri):
                 continue
 
             schema = {"props": {}, "subject": None, "sm_node_id": u.id}
-            for us_edge in sm.outgoing_edges(u.id):
+            for us_edge in sm.out_edges(u.id):
                 if us_edge.abs_uri in self.ID_PROPS:
                     v = sm.get_node(us_edge.target)
-                    assert v.is_data_node
+                    assert isinstance(v, O.DataNode)
                     assert schema["subject"] is None
                     schema["subject"] = v.col_index
                     continue
 
                 s = sm.get_node(us_edge.target)
-                assert s.is_class_node and WDOnt.is_uri_statement(s.abs_uri)
-                assert WDOnt.is_uri_property(us_edge.abs_uri)
+                assert isinstance(s, O.ClassNode) and wdns.is_abs_uri_statement(
+                    s.abs_uri
+                )
+                assert wdns.is_abs_uri_property(us_edge.abs_uri)
 
-                pnode = WDOnt.get_prop_id(us_edge.abs_uri)
+                pnode = wdns.get_prop_id(us_edge.abs_uri)
                 if pnode not in schema["props"]:
                     schema["props"][pnode] = []
 
@@ -598,44 +558,44 @@ class WikidataSemanticModelHelper(WDOnt):
                     "qualifiers": [],
                 }
                 schema["props"][pnode].append(stmt)
-                for sv_edge in sm.outgoing_edges(s.id):
+                for sv_edge in sm.out_edges(s.id):
                     v = sm.get_node(sv_edge.target)
 
-                    assert WDOnt.is_uri_property(sv_edge.abs_uri)
+                    assert wdns.is_abs_uri_property(sv_edge.abs_uri)
                     if sv_edge.abs_uri == us_edge.abs_uri:
                         assert stmt["value"] is None, "only one property"
                         # this is property
-                        if v.is_class_node:
+                        if isinstance(v, O.ClassNode):
                             stmt["value"] = {"type": "classnode", "value": v.id}
-                        elif v.is_data_node:
+                        elif isinstance(v, O.DataNode):
                             stmt["value"] = {"type": "datanode", "value": v.col_index}
                         else:
-                            assert v.is_literal_node
+                            assert isinstance(v, O.LiteralNode)
                             stmt["value"] = {"type": "literalnode", "value": v.value}
                     else:
                         # this is qualifier
-                        if v.is_class_node:
+                        if isinstance(v, O.ClassNode):
                             stmt["qualifiers"].append(
                                 {
                                     "type": "classnode",
-                                    "pnode": WDOnt.get_prop_id(sv_edge.abs_uri),
+                                    "pnode": wdns.get_prop_id(sv_edge.abs_uri),
                                     "value": v.id,
                                 }
                             )
-                        elif v.is_data_node:
+                        elif isinstance(v, O.DataNode):
                             stmt["qualifiers"].append(
                                 {
                                     "type": "datanode",
-                                    "pnode": WDOnt.get_prop_id(sv_edge.abs_uri),
+                                    "pnode": wdns.get_prop_id(sv_edge.abs_uri),
                                     "value": v.col_index,
                                 }
                             )
                         else:
-                            assert v.is_literal_node
+                            assert isinstance(v, O.LiteralNode)
                             stmt["qualifiers"].append(
                                 {
                                     "type": "literalnode",
-                                    "pnode": WDOnt.get_prop_id(sv_edge.abs_uri),
+                                    "pnode": wdns.get_prop_id(sv_edge.abs_uri),
                                     "value": v.value,
                                 }
                             )
@@ -677,8 +637,8 @@ class WikidataSemanticModelHelper(WDOnt):
                     ents = [
                         {
                             "id": qnode_id,
-                            "uri": WDOnt.get_qnode_uri(qnode_id),
-                            "label": self.qnodes[qnode_id].label,
+                            "uri": wdns.get_entity_abs_uri(qnode_id),
+                            "label": self.wdentities[qnode_id].label,
                             "props": {},
                         }
                         for qnode_id in qnode_ids
@@ -786,3 +746,19 @@ class WikidataSemanticModelHelper(WDOnt):
                     new_record[ent_id] = ent
             new_records.append(new_record)
         return records
+
+    def get_qnode_label(self, qid: str):
+        """Get WDEntity label from id"""
+        if qid in self.wdclasses:
+            label = self.wdclasses[qid].label
+        elif qid in self.wdentities:
+            label = self.wdentities[qid].label
+        else:
+            label = self.wdentity_labels.get(qid, qid)
+        return f"{label} ({qid})"
+
+    def get_pnode_label(self, pid: str):
+        """Get PNode label from id"""
+        if pid not in self.wdprops:
+            return pid
+        return f"{self.wdprops[pid].label} ({pid})"
